from __future__ import annotations

# Kept here for backwards compatibility
from typing import Any, Optional, Protocol, TypeVar

from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser
from django.http import HttpRequest, HttpResponse
from django.utils.module_loading import import_string

R_ = TypeVar("R_", bound=HttpResponse)


class ResponseHookProtocol(Protocol):
    def __call__(
        self,
        request: HttpRequest,
        response: R_,
        user: AbstractBaseUser,
        cognito_jwt: dict[str, Any],
    ) -> R_:
        ...


class Settings:
    """
    Shadow Django's settings with a little logic
    """

    @property
    def COGNITO_CONFIG(self) -> dict:
        return getattr(settings, "COGNITO_CONFIG", {})

    @property
    def COGNITO_ENDPOINT(self) -> str:
        return self.COGNITO_CONFIG.get("ENDPOINT", "")

    @property
    def COGNITO_AUTH_ENDPOINT(self) -> str:
        return f"{self.COGNITO_ENDPOINT}/oauth2/authorize"

    @property
    def COGNITO_TOKEN_ENDPOINT(self) -> str:
        return f"{self.COGNITO_ENDPOINT}/oauth2/token"

    @property
    def COGNITO_CLIENT_ID(self) -> str:
        return self.COGNITO_CONFIG.get("CLIENT_ID", "")

    @property
    def COGNITO_CLIENT_SECRET(self) -> str:
        return self.COGNITO_CONFIG.get("CLIENT_SECRET", "")

    @property
    def COGNITO_REDIRECT_URI(self) -> Optional[str]:
        return self.COGNITO_CONFIG.get("REDIRECT_URI", "")

    @property
    def COGNITO_JWKS_URI(self) -> str:
        return self.COGNITO_CONFIG.get("JWKS_URI", "")

    @property
    def COGNITO_REQUIRED_GROUPS(self) -> list[str]:
        return self.COGNITO_CONFIG.get("REQUIRED_GROUPS", [])

    @property
    def response_hook(self) -> ResponseHookProtocol:
        fn = self.COGNITO_CONFIG.get("RESPONSE_HOOK", "")
        return import_string(fn) if fn else default_response_hook


def default_response_hook(
    request: HttpRequest,
    response: HttpResponse,
    user: AbstractBaseUser,
    cognito_jwt: dict[str, Any],
) -> HttpResponse:
    return response


COGNITO_GROUPS_SESSION_KEY = "COGNITO_GROUPS"
conf = Settings()



from logging import getLogger
from typing import Any, Optional

from django.contrib.auth.backends import RemoteUserBackend
from django.contrib.auth.models import AbstractBaseUser
from django.http import HttpRequest

from django_cognito_saml.conf import COGNITO_GROUPS_SESSION_KEY, conf

logger = getLogger("django_cognito_saml")


class CognitoUserBackend(RemoteUserBackend):
    """
    This backend is to be used in conjunction with the ``RemoteUserMiddleware``
    found in the middleware module of this package, and is used when the server
    is handling authentication outside of Django.

    By default, the ``authenticate`` method creates ``User`` objects for
    usernames that don't already exist in the database.  Subclasses can disable
    this behavior by setting the ``create_unknown_user`` attribute to
    ``False``.

    Override configure_user for post login customization

    """

    # Change this to False if you do not want to create a remote user.
    create_unknown_user = True

    cognito_jwt: dict

    def authenticate(  # type: ignore[override]
        self, request: HttpRequest, cognito_jwt: dict[str, Any], **kwargs: Any
    ) -> Optional[AbstractBaseUser]:
        self.cognito_jwt = cognito_jwt
        remote_user = cognito_jwt["email"]
        return super().authenticate(request, remote_user=remote_user, **kwargs)

    def configure_user(  # type: ignore[override]
        self, request: HttpRequest, user: AbstractBaseUser, **kwargs: Any
    ) -> AbstractBaseUser:
        """
        Configure a user after creation and return the updated user.

        By default, return the user unmodified.
        """
        return user


class SuperUserBackend(CognitoUserBackend):
    """
    This backend creates users and adds them as superusers
    """

    # Change this to False if you do not want to create a remote user.
    create_unknown_user = True

    cognito_jwt: dict

    def authenticate(  # type: ignore[override]
        self, request: HttpRequest, cognito_jwt: dict[str, Any], **kwargs: Any
    ) -> Optional[AbstractBaseUser]:
        url = request.build_absolute_uri("/")

        # Extracting email
        email = cognito_jwt.get("email")

        # Parsing email
        unparsed_groups = cognito_jwt.get("custom:groups", "")
        if "[" in unparsed_groups:
            groups = unparsed_groups.removeprefix("[").removesuffix("]").split(",")
        else:
            groups = [unparsed_groups.strip()]

        groups += cognito_jwt.get("cognito:groups", [])

        # Checking if user has access.
        if conf.COGNITO_REQUIRED_GROUPS:
            if not set(groups).issuperset(set(conf.COGNITO_REQUIRED_GROUPS)):
                msg = (
                    f"COGNITO_SSO_FAILURE: {email} logged into "
                    f"{url} but was not in groups: {conf.COGNITO_REQUIRED_GROUPS}"
                )
                logger.info(msg)
                return None

        user = super().authenticate(request, cognito_jwt=cognito_jwt, **kwargs)

        if user:
            request.session[COGNITO_GROUPS_SESSION_KEY] = groups
            msg = (
                f"COGNITO_SSO_LOGIN: {email} logged into "
                f"{url} with groups: {groups}"
            )
            logger.info(msg)

        return user

    def configure_user(  # type: ignore[override]
        self,
        request: HttpRequest,
        user: AbstractBaseUser,
        created: bool = True,
        **kwargs: Any,
    ) -> AbstractBaseUser:
        """
        Configure a user after creation and return the updated user.

        By default, return the user unmodified.
        """

        if created:
            user.set_unusable_password()
            user.email = self.cognito_jwt["email"]  # type: ignore [attr-defined]
            user.first_name = self.cognito_jwt.get("given_name") or ""  # type: ignore
            user.last_name = self.cognito_jwt.get("family_name") or ""  # type: ignore
            user.is_superuser = True  # type: ignore [attr-defined]
            user.is_staff = True  # type: ignore [attr-defined]
            user.save()
        return user


class DisallowUnknownUserSuperUserBackend(SuperUserBackend):
    """Same backend as `SuperUserBackend` but does not create
    users if they don't already exist.
    """

    create_unknown_user = False



SECRET_KEY = "asdasd"

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": "mem_db"}}


INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django_cognito_saml",
]

MIDDLEWARE = ["django.contrib.sessions.middleware.SessionMiddleware"]


ROOT_URLCONF = "django_cognito_saml.urls"


COGNITO_CONFIG = dict(
    CLIENT_ID="client_id",
    CLIENT_SECRET="client_secret",
    ENDPOINT="https://fake_cognito.com",
    JWKS_URI="http://cognito-idp.ap-southeast-2.amazonaws.com:443/ap-southeast-some_endpoint/.well-known/jwks.json",  # noqa
    REDIRECT_URI="https://test_redirect.com/",
    RESPONSE_HOOK="",
)

AUTHENTICATION_BACKENDS = ("django_cognito_saml.backends.CognitoUserBackend",)



import base64
import json
from typing import Any, cast
from urllib.parse import parse_qsl, urlsplit
from uuid import uuid4

from django.conf import settings
from django.http import HttpRequest
from django.urls import NoReverseMatch, reverse


def generate_state(request: "HttpRequest") -> str:
    return base64.b64encode(
        json.dumps(
            {
                "redirect_uri": get_cognito_callback_url(request),
                "key": str(uuid4()),
            }
        ).encode()
    ).decode()


def generate_return_uri(request: "HttpRequest") -> str:
    """Try to use `next` url if passed through
    if not default to settings.LOGIN_REDIRECT_URL"""
    referrer = cast(str, request.META.get("HTTP_REFERER"))
    if not referrer:
        return settings.LOGIN_REDIRECT_URL

    referrer_path = urlsplit(referrer)[2]
    if referrer_path == _reverse_or_blank("login") or referrer_path.startswith(
        _reverse_or_blank("rest_framework:login")
    ):
        # If referred from login page, try to grab the ?next param.
        return (
            dict(parse_qsl(urlsplit(referrer).query)).get("next")
            or settings.LOGIN_REDIRECT_URL
        )
    elif referrer_path.startswith(
        _reverse_or_blank("logout")
    ) or referrer_path.startswith(_reverse_or_blank("admin:logout")):
        # If referrer is logout page, don't take them back there..
        return settings.LOGIN_REDIRECT_URL
    else:
        # Else just redirect back to referrer.
        return referrer


def get_cognito_callback_url(request: HttpRequest) -> str:
    return request.build_absolute_uri(reverse("cognito_callback"))


def _reverse_or_blank(url: str, *args: Any, **kwargs: Any) -> str:
    try:
        return reverse(url, *args, **kwargs)
    except NoReverseMatch:
        return "____IMPOSSIBLY_BLANK_URL"



from typing import ClassVar, List

from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin

from .conf import COGNITO_GROUPS_SESSION_KEY


class CognitoPermissionMixin(LoginRequiredMixin, UserPassesTestMixin):
    """This mixin tests a user is in a required cognito group"""

    groups_required: ClassVar[List[str]] = []

    def test_func(self) -> bool:
        if request := getattr(self, "request", None):
            breakpoint()
            cognito_perms = request.session.get(COGNITO_GROUPS_SESSION_KEY, [])
            assert isinstance(cognito_perms, list)
            for group in self.groups_required:
                if group not in cognito_perms:
                    return False
            return True
        return False



import json
import time
import urllib
from typing import Dict, Optional
from urllib.parse import urlencode

import jwt
from django.conf import settings
from django.contrib import auth
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect, JsonResponse

from .conf import conf
from .utils import generate_return_uri, generate_state, get_cognito_callback_url


def cognito_login(request: "HttpRequest") -> "HttpResponse":
    """Redirect users to the cognito hosted login view.

    Encoded within the state is a redirect (utilizing the bouncer).

    Post login; users are directed to "cognito_callback"
    """
    state = generate_state(request)

    params = {
        "client_id": conf.COGNITO_CLIENT_ID,
        "response_type": "code",
        "scope": "openid",
        "redirect_uri": conf.COGNITO_REDIRECT_URI or get_cognito_callback_url(request),
        "state": state,
    }

    request.session["state"] = state
    request.session["next"] = generate_return_uri(request)

    redirect_url = f"{conf.COGNITO_AUTH_ENDPOINT}?{urlencode(params)}"
    return HttpResponseRedirect(redirect_url)


JWKS_CLIENT = jwt.PyJWKClient(
    conf.COGNITO_JWKS_URI, cache_keys=True, cache_jwk_set=True, lifespan=60 * 60
)


def _decode_jwt(token: str) -> Dict:
    # Fetch JWT token via authorization code flow
    jwt_signing_key = JWKS_CLIENT.get_signing_key_from_jwt(token)

    return jwt.decode(
        token,
        jwt_signing_key.key,
        algorithms=["RS256"],
        audience=conf.COGNITO_CLIENT_ID,
    )


def _exchange_code_for_jwt(request: HttpRequest, code: str) -> Optional[Dict]:
    # Fetch JWT token via authorization code flow
    # Creating a payload to fetch JWT tokens from cognito
    data = {
        "code": code,
        "client_id": conf.COGNITO_CLIENT_ID,
        "client_secret": conf.COGNITO_CLIENT_SECRET,
        "redirect_uri": conf.COGNITO_REDIRECT_URI or get_cognito_callback_url(request),
        "grant_type": "authorization_code",
    }

    encoded_data = urllib.parse.urlencode(data).encode("utf-8")

    cognito_request = urllib.request.Request(conf.COGNITO_TOKEN_ENDPOINT, encoded_data)
    cognito_response = urllib.request.urlopen(cognito_request)

    status_code = cognito_response.status

    if status_code != 200:
        return None

    return json.loads(cognito_response.read())


def cognito_callback(request: "HttpRequest") -> "HttpResponse":
    if not request.GET.get("code"):
        return JsonResponse({"detail": "Code parameter missing"}, status=401)

    if not request.GET.get("state"):
        return JsonResponse({"detail": "State parameter missing"}, status=401)

    if request.GET.get("state") != request.session.get("state"):
        return JsonResponse({"detail": "Invalid state parameter"}, status=401)

    cognito_data = _exchange_code_for_jwt(request, code=request.GET["code"])
    if not cognito_data:
        return JsonResponse({"detail": "Invalid token response"}, status=401)

    try:
        decoded_jwt_token = _decode_jwt(cognito_data["id_token"])
    except jwt.ImmatureSignatureError:
        # The token provided by AWS requires a second to be active :(
        time.sleep(1)
        decoded_jwt_token = _decode_jwt(cognito_data["id_token"])

    # Authenticate user
    user = auth.authenticate(request=request, cognito_jwt=decoded_jwt_token)

    if not user:
        return JsonResponse({"detail": "User account not found"}, status=404)

    auth.login(request, user)

    # Redirect
    redirect = request.session.get("next", None)
    redirect_default = settings.LOGIN_REDIRECT_URL
    response = HttpResponseRedirect(redirect or redirect_default)

    # Post response hook
    response = conf.response_hook(
        request, response, user=user, cognito_jwt=decoded_jwt_token
    )
    return response



from django.urls import path

from . import views

urlpatterns = [
    path("login/", views.cognito_login, name="cognito_login"),
    path("cognito/callback/", views.cognito_callback, name="cognito_callback"),
]
